@@ -9,7 +9,10 @@ from ..errors import (
     InvalidParameter,
     RetriesExceeded,
     CookieError,
-    UnexpectedError
+    UnexpectedError,
+    URLNotProvided,
+    SiteNotSupported,
+    InvalidURL
 )
 
 from ..utils.core import (
@@ -18,7 +21,10 @@ from ..utils.core import (
     safe_print
 )
 
-from ..utils.timed_utils import timed_input
+from ..utils.timed_utils import (
+    timed_input,
+    interruptible_sleep
+)
 from ..debugging import log
 
 
@@ -223,6 +229,9 @@ class Chat():
         # TODO
         # author/user/uploader/creator
 
+        self._output_writer = None
+        self._output_callback = None
+
     def __iter__(self):
         """Allows the object to be iterable
 
@@ -231,25 +240,44 @@ class Chat():
         """
         return self
 
+    def attach_writer(self, writer):
+        # writer is a ContinuousWriter
+        self._output_writer = writer
+        if self._output_writer.is_default():
+            self._output_callback = lambda item: self._output_writer.write(
+                self.format(item), flush=True)
+        else:
+            self._output_callback = lambda item: self._output_writer.write(
+                item, flush=True)
+
     def __next__(self):
         """Get the next chat message from the generator
 
         :return: The next chat item
         :rtype: dict
         """
-        item = next(self.chat)
-        if self.callback:
-            self.callback(item)
-        return item
+        try:
+            item = next(self.chat)
+            if self.callback:  # user-defined callback
+                self.callback(item)
 
-    def print_formatted(self, item):
+            if self._output_callback:  # output callback
+                self._output_callback(item)
+
+            return item
+        except StopIteration as e:
+            # Safely close output file when done
+            if self._output_writer is not None:
+                self._output_writer.close()
+            raise e
+
+    def print_formatted(self, item, flush=True):
         """Safely print the formatted message
 
         :param item: The chat item to be printed
         :type item: dict
         """
-        formatted = self.format(item)
-        safe_print(formatted)
+        safe_print(self.format(item), flush=flush)
 
     def format(self, item):
         """Format chat messages
@@ -278,6 +306,14 @@ class BaseChatDownloader:
 
     # For general tests (non-site specific)
     _TESTS = [
+        {
+            'name': 'Inactivity timeout',
+            'params': {
+                'url': 'https://twitch.tv/xenova',
+                'inactivity_timeout': 5,
+                'timeout': 20,  # As a fallback
+            }
+        },
         {
             'name': 'Get a certain number of messages from a livestream.',
             'params': {
@@ -289,6 +325,48 @@ class BaseChatDownloader:
             'expected_result': {
                 'messages_condition': lambda messages: len(messages) <= 10,
             }
+        },
+
+
+        {
+            'name': 'Scheme not supplied',
+            'params': {
+                'url': 'www.youtube.com/watch?v=5qap5aO4i9A',
+                'max_messages': 10,
+                'timeout': 60,  # As a fallback
+            },
+            'expected_result': {
+                'messages_condition': lambda messages: len(messages) <= 10,
+            }
+        },
+
+        # Tests for errors
+        {
+            'name': 'No URL provided.',
+            'params': {
+                'url': '',
+            },
+            'expected_result': {
+                'error': URLNotProvided
+            }
+        },
+        {
+            'name': 'Site not supported',
+            'params': {
+                'url': 'https://www.example.com',
+            },
+            'expected_result': {
+                'error': SiteNotSupported
+            }
+        },
+        {
+            'name': 'Invalid URL',
+            'params': {
+                'url': '#',
+            },
+            'expected_result': {
+                'error': InvalidURL
+            }
         }
     ]
 
@@ -483,13 +561,13 @@ class BaseChatDownloader:
         return new_dict
 
     @staticmethod
-    def retry(attempt_number, max_attempts, error=None, retry_timeout=None, text=None):
+    def retry(attempt_number, max_attempts=1, error=None, retry_timeout=None, text=None, interruptible_retry=True, **kwargs):
         """Retry to occur after an error occurs
 
         :param attempt_number: The current attempt number
         :type attempt_number: int
         :param max_attempts: The maximum number of attempts allowed
-        :type max_attempts: int
+        :type max_attempts: int, optional
         :param error: The error which was raised, defaults to None
         :type error: Exception, optional
         :param retry_timeout: The number of seconds to sleep after failing,
@@ -521,7 +599,11 @@ class BaseChatDownloader:
 
         must_sleep = time_to_sleep >= 0
         if must_sleep:
-            sleep_text = '(sleep for {}s or press Enter)'.format(time_to_sleep)
+            if interruptible_retry:
+                sleep_text = '(sleep for {}s or press Enter)'.format(
+                    time_to_sleep)
+            else:
+                sleep_text = '(sleep for {}s)'.format(time_to_sleep)
         else:
             sleep_text = ''
 
@@ -530,11 +612,6 @@ class BaseChatDownloader:
         if isinstance(error, Exception):
             retry_text += ' {} ({})'.format(error, error.__class__.__name__)
 
-        log(
-            'warning',
-            text + [retry_text]
-        )
-
         if isinstance(error, JSONDecodeError):
             log(
                 'debug',
@@ -544,8 +621,16 @@ class BaseChatDownloader:
             if page_title:
                 log('debug', 'Title: {}'.format(page_title))
 
+        log(
+            'warning',
+            text + [retry_text]
+        )
+
         if must_sleep:
-            timed_input(time_to_sleep)
+            if interruptible_retry:
+                timed_input(time_to_sleep)
+            else:
+                interruptible_sleep(time_to_sleep)
         else:
             pause()
 
